import re

from pyjo_mdl.exceptions import InstanceValidationError, ModelValidationError

__author__ = 'xelhark'
URL_REGEX = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')


def number_validator(fs, value):
    if 'max_value' in fs and value > fs['max_value']:
        raise InstanceValidationError(
            "Value is bigger than max value: {} > {}".format(value, fs['max_value']))
    if 'min_value' in fs and value < fs['min_value']:
        raise InstanceValidationError(
            "Value is smaller than min value: {} < {}".format(value, fs['min_value']))


def string_validator(fs, value):
    if 'max_length' in fs and len(value) > fs['max_length']:
        raise InstanceValidationError("Value {} is longer than max allowed length ({})".format(value, fs['max_length']))
    if 'validation' in fs and not re.match(fs['validation'], value):
        raise InstanceValidationError("Value {} does not match provided validation regex".format(value))
    if 'values' in fs and value not in fs['values']:
        raise InstanceValidationError(
            "Value {} does not match any value in the enum list [{}]".format(value, ', '.join(fs['values']))
        )


def url_validator(fs, value):
    if not URL_REGEX.match(value):
        raise InstanceValidationError("{} is not a valid URL.".format(value))


def embedded_validator(field_structure, *args, **kwargs):
    from pyjo_mdl.core import pyjo_model_from_structure
    return pyjo_model_from_structure(structure=field_structure['model'], *args, **kwargs).validate


def cast_bool(value):
    if isinstance(value, str):
        if value.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif value.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
    raise ValueError('Boolean value expected. Got "{}" instead'.format(value))



from pyjo import Model
from pyjo.exceptions import FieldTypeError, RequiredFieldError
from six import iteritems

from pyjo_mdl.exceptions import InstanceValidationError

__author__ = 'xelhark'


class ValidatableModel(Model):

    @classmethod
    def validate(cls, values):
        for key, field in iteritems(cls._fields):
            field.cast_and_validate(values[key])


class FullErrorValidation(ValidatableModel):
    def __init__(self, **kwargs):
        errors = []
        for key, value in iteritems(kwargs):
            try:
                self._fields[key].cast_and_validate(value)
            except InstanceValidationError as err:
                if err.errors_list:
                    errors.extend(err.errors_list)
                else:
                    err = InstanceValidationError("Error validating field \'{}\': {}".format(key, str(err)))
                    errors.append(err)

        if errors:
            raise InstanceValidationError("Error validating instance.", errors_list=errors)
        super(FullErrorValidation, self).__init__(**kwargs)

    @classmethod
    def validate(cls, values):
        errors = []
        for key, field in iteritems(cls._fields):
            try:
                field.cast_and_validate(values.get(key))
            except (InstanceValidationError, FieldTypeError, RequiredFieldError) as err:
                err = InstanceValidationError("Error validating field \'{}\': {}".format(key, str(err)))
                errors.append(err)
        if errors:
            raise InstanceValidationError("Error validating instance.", errors_list=errors)



__author__ = 'xelhark'


class ErrorWithList(ValueError):
    def __init__(self, message, errors_list=None):
        self.errors_list = errors_list
        super(ErrorWithList, self).__init__(message)


class ModelValidationError(ErrorWithList):
    pass


class InstanceValidationError(ErrorWithList):
    pass



import copy
from typing import Type

from pyjo import Field, ModelMetaclass, ListField
from six import iteritems

from pyjo_mdl.helpers import number_validator, cast_bool, string_validator, url_validator, embedded_validator
from pyjo_mdl.model_classes import ValidatableModel
from .exceptions import ModelValidationError

__author__ = 'xelhark'

BASE_FIELD_PROPERTIES = {
    'integer': {
        'type': int,
        'model_validator': None,
        'validator': number_validator,
    },
    'float': {
        'type': float,
        'model_validator': None,
        'validator': number_validator,
    },
    'boolean': {
        'type': bool,
        'cast': cast_bool,
    },
    'string': {
        'type': str,
        'model_validator': None,
        'validator': string_validator,
    },
    'url': {
        'type': str,
        'model_validator': None,
        'validator': url_validator,
    },
    'embedded': {
        'type': None,
        'validator_factory': embedded_validator,
    },
    'array': {
        'type': None,
        'field': ListField,
        'field_kwargs': {
            'inner_field': lambda fs, *args, **kwargs: field_from_structure(fs['element'], *args, **kwargs),
        },
        'model_validator': None,
    },
}


def field_from_structure(fs, fields_definitions, base_class=None):
    if not isinstance(fs, dict):
        raise ModelValidationError("Invalid model. It should be a dictionary with a \'type\' property.")
    if 'type' not in fs:
        raise ModelValidationError("No \'type\' property found.")

    if fs['type'] not in fields_definitions:
        raise ModelValidationError("Unsupported field type: {}".format(fs['type']))

    props = fields_definitions[fs['type']]

    vf = props.get('validator_factory')
    if vf:
        validator = vf(field_structure=fs, fields_properties=fields_definitions, base_class=base_class)
    else:
        validator = lambda value: props.get('validator', lambda *args, **kwargs: None)(fs, value)

    field_cls = props.get('field', Field)
    field_kwargs = dict(
        type=props.get('type'),
        cast=props.get('cast'),
        validator=validator,
        required=fs.get('required', True),
    )
    if field_cls != Field and 'field_kwargs' in props:
        del field_kwargs['type']
        other_field_kwargs = {}
        for k, v in iteritems(props.get('field_kwargs')):
            if callable(v):
                v = v(fs, fields_definitions, base_class=base_class)
            other_field_kwargs[k] = v
        field_kwargs.update(other_field_kwargs)

    return field_cls(**field_kwargs)


T = Type[ValidatableModel]


def pyjo_model_from_structure(
        structure: dict,
        base_class: T = ValidatableModel,
        fields_properties: dict = None
) -> T:
    fields = {}
    errors = []

    fields_definitions = copy.deepcopy(BASE_FIELD_PROPERTIES)
    if fields_properties:
        fields_definitions.update(fields_properties)

    for element_key, field_structure in structure.items():
        if element_key in fields:
            error = ModelValidationError("Duplicate field definition: {}".format(element_key))
            errors.append(error)

        if element_key[0] == '_':
            error = ModelValidationError("First character of field cannot be '_': {}".format(element_key))
            errors.append(error)

        try:
            fields[element_key] = field_from_structure(field_structure, fields_definitions, base_class=base_class)
        except ModelValidationError as err:
            errors.append(err)

    if errors:
        raise ModelValidationError("Invalid model", errors_list=errors)
    return ModelMetaclass.__new__(ModelMetaclass, "Validator", (base_class,), fields)



#!/usr/bin/env python
from setuptools import setup, find_packages
import re

with open('pyjo_mdl/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)
if not version:
    raise RuntimeError('Cannot find version information')

setup(
    name='pyjo-mdl',
    version=version,
    description='Model Definition Language for Pyjo',
    url='https://github.com/xelhark/pyjo-mdl',
    long_description=open('README.md').read(),
    author='Gabriele Platania',
    author_email='gabriele.platania@gmail.com',
    packages=find_packages(),  # ['pyjo_mdl'],
    package_data={'': ['LICENSE']},
    install_requires=[
        'pyjo'
    ],
)
